'' =================================================================================================
''
''   File....... efx-tek_ouija_board__2017-10-31.spin
''   Purpose....
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2011-2017 Jon McPhalen
''   E-mail.....
''   Started....
''   Updated.... 31 NOV 2017
''
'' =================================================================================================

{

  To change text color for spelling:

  - T-RED
  - T-ORANGE
  - T-YELLOW
  - T-GREEN
  - T-CYAN
  - T-BLUE
  - T-VIOLET
  - T-PURPLE
  - T-PINK
  - T-WHITE

  To change board background color:

  - B-RED
  - B-ORANGE
  - B-YELLOW
  - B-GREEN
  - B-CYAN
  - B-BLUE
  - B-VIOLET
  - B-PURPLE
  - B-PINK
  - B-WHITE

  Non-spelled responses:

  - YES  { green }
  - NO   { red   }
  - BYE  { white }

}


con

  ' Important: This program is designed to be used on Rev F and later boards
  '            -- requires mods for Rev E or lower

  PCB_REV = "F"


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = _xinfreq * (_clkmode >> 6)                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us

  TRM_BAUD = 115_200                                            ' terminal baud rate
  BLE_BAUD =   9_600                                            ' ble baud rate


con { io pins }

  RX1     = 31  { I }                                           ' serial / programming
  TX1     = 30  { O }

  SDA     = 29  { I/O }                                         ' i2c / eeprom
  SCL     = 28  { I/O }

  XP0     = 27  { I/O }                                         ' extra port 0 - not on Rev E or lower
  XP1     = 24  { I/O }                                         ' extra port 1 - not on Rev E or lower
  XP2     = 19  { I/O }                                         ' extra port 2 - not on Rev E or lower
  XP3     = 16  { I/O }                                         ' extra port 3 - not on Rev E or lower

  SIO_RX  = 26  { I/O }                                         ' TTL serial io
  SIO_TX  = 25  { I/O }                                         ' TTL serial io (not enabled)

  RX2     = 23  { I }                                           ' RS-485 port
  RX2_EN  = 22  { O }
  TX2_EN  = 21  { O }
  TX2     = 20  { O }

  G_LED   = 18  { O }                                           ' bi-color LED (Rev F assignments)
  R_LED   = 17  { O }

  OUT7    = 15  { O }                                           ' outputs
  OUT6    = 14  { O }
  OUT5    = 13  { O }
  OUT4    = 12  { O }
  OUT3    = 11  { O }
  OUT2    = 10  { O }
  OUT1    =  9  { O }
  OUT0    =  8  { O }

  OPT_BR  =  7  { I/O }                                         ' options switches
  OPT_A1  =  6  { I/O }
  OPT_A0  =  5  { I/O }
  OPT_SM  =  4  { I/O }

  DMX_A8  =  3  { I }                                           ' high-bit of DMX address

  DO_165  =  2  { I }                                           ' control lines for 74x165
  CLK_165 =  1  { O }
  LD_165  =  0  { O }


con { app connections }

  RX3     = OPT_A0  { I }                                       ' for Bluetooth module
  TX3     = OPT_SM  { O }


con { serial connection for remote commands }

  #0, M_USB, M_RADIO                                            ' serial connection modes

  CMD_MODE = M_RADIO 'M_USB                                              ' command mode


con

  CHANNELS = 7                                                  ' protect OUT7 (for WS28xx pixels)
  LAST_CH  = CHANNELS - 1

  STRIP_LEN = 64                                                ' # of bulbs in strip

  FADE_UP_MS = 2000                                             ' fade controls for spelling
  FADE_DN_MS = 1000

  #(true), ON, OFF                                              ' digital control

  #0, NO, YES

  #0, BLK, RED, GRN, YEL                                        ' for r/g led

  #0, NOCON, RX, TX, RXTX                                       ' for rs-485

  #0, LSBFIRST, MSBFIRST

  BUF_SIZE = 40


obj

' main                                                          ' * master Spin cog
  time       : "jm_time_80"                                     '   timing and delays (80MHz system)
  prng       : "jm_prng"                                        '   pseudo-random number generator
  tbuf       : "jm_textbuf"                                     '   text buffer manipulation
  io         : "jm_io"                                          '   essential io
  outs       : "jm_pwm8"                                        ' * pwm driver for outputs
  cmd        : "jm_fullduplexserial"                            ' * serial IO for ble or terminal
  strip      : "jm_rgbx_pixel"                                  ' * ws2812 control
' background                                                    ' * background RGY LED control, inputs scanning

' * uses cog when loaded


var

  long  pixels[STRIP_LEN]                                       ' pixel buffer

  long  bcolor                                                  ' background color
  long  tcolor                                                  ' text color

  byte  buffer[BUF_SIZE+1]                                      ' incoming text buffer


dat

  OB_Red                long    $80_00_00_00
  OB_Orange             long    $50_10_00_00
  OB_Yellow             long    $50_38_00_00
  OB_Green              long    $00_80_00_00
  OB_Cyan               long    $00_42_42_00
  OB_Blue               long    $00_00_80_00
  OB_Violet             long    $40_00_40_00
  OB_Purple             long    $30_00_60_00
  OB_White              long    $2A_2A_2A_00
  OB_Pink               long    $60_20_20_00

  ' positions of letters in string display
  ' -- table contains WS28xx channel # for letter

                        byte

  OB_Yes                byte     5
  OB_No                 byte     11

                        '        a   b   c   d   e   f   g   h   i   j   k   l   m
  Alphas  { A..Z }      byte     0,  1,  2,  3,  6,  7,  8,  9, 10, 13, 14, 15, 16
                        '        n   o   p   q   r   s   t   u   v   w   x   y   z
                        byte    29, 28, 27, 26, 25, 24, 23, 22, 21, 20, 19, 18, 17

                        '        0   1   2   3   4   5   6   7   8   9
  Numbers { 0..9 }      byte    41, 31, 32, 33, 34, 35, 36, 37, 39, 40

  OB_GoodbyeLo          byte    43                              ' first bulb in group
  OB_GoodbyeHi          byte    49                              ' last bulb in group


pub main

  setup                                                         ' setup io and objects
  cmd.rxflush                                                   ' clean-up

  bcolor := OB_Yellow
  tcolor := OB_Yellow

  fade_up_group(0, STRIP_LEN-1, bcolor, FADE_UP_MS)

  repeat
    get_string
    check_commands
    if (tbuf.size)                                              ' unknown text in buffer?
      fade_down_group(0, STRIP_LEN-1, bcolor, FADE_DN_MS)
      time.pause(500)
      show_string
      time.pause(1000)
      fade_up_group(0, STRIP_LEN-1, bcolor, FADE_UP_MS)


pub fade_up_group(lo, hi, color, ms) | tix, t, level, c

  ms #>= 100

  tix := ms * MS_001 / 255

  t := cnt
  repeat level from 1 to 255
    c := strip.scale_rgbw(color, strip.gamma(level))
    longfill(@pixels[lo], c, hi-lo)
    waitcnt(t += tix)


pub fade_down_group(lo, hi, color, ms) | tix, t, level, c

  ms #>= 100                                                   '

  tix := ms * MS_001 / 255

  t := cnt
  repeat level from 254 to 0
    c := strip.scale_rgbw(color, strip.gamma(level))
    longfill(@pixels[lo], c, hi-lo)
    waitcnt(t += tix)


pub fade_up(ch, color, ms) | tix, t, level

  tix := ms * MS_001 / 255

  t := cnt
  repeat level from 1 to 255
    strip.setx(ch, color, strip.gamma(level))
    waitcnt(t += tix)


pub fade_down(ch, color, ms) | tix, t, level

  tix := ms * MS_001 / 255

  t := cnt
  repeat level from 254 to 0
    strip.setx(ch, color, strip.gamma(level))
    waitcnt(t += tix)


pub get_string | c

'' Get command or string from serial input stream

  tbuf.clear

  repeat until (cmd.available)                                  ' wait for anything in buffer

  repeat
    if (CMD_MODE == M_USB)
      c := tbuf.upper(cmd.rx)                                   ' using terminal (PST)
    else
      c := tbuf.upper(cmd.rxtime(2))                            ' from BT/BLE

    case c
      8 :
        tbuf.backspace

      -1, 13 :                                                  ' -1 for timeout, 13 for Enter
        return

      " ", "-", "A".."Z", "0".."9" :
        if (tbuf.size < BUF_SIZE)
          tbuf.putc(c)


dat { commands }

  Commands      byte

  fgRed         byte    "T-RED", 0
  fgOrange      byte    "T-ORANGE", 0
  fgYellow      byte    "T-YELLOW", 0
  fgGreen       byte    "T-GREEN", 0
  fgCyan        byte    "T-CYAN", 0
  fgBlue        byte    "T-BLUE", 0
  fgViolet      byte    "T-VIOLET", 0
  fgPurple      byte    "T-PURPLE", 0
  fgPink        byte    "T-PINK", 0
  fgWhite       byte    "T-WHITE", 0

  bgRed         byte    "B-RED", 0
  bgOrange      byte    "B-ORANGE", 0
  bgYellow      byte    "B-YELLOW", 0
  bgGreen       byte    "B-GREEN", 0
  bgCyan        byte    "B-CYAN", 0
  bgBlue        byte    "B-BLUE", 0
  bgViolet      byte    "B-VIOLET", 0
  bgPurple      byte    "B-PURPLE", 0
  bgPink        byte    "B-PINK", 0
  bgWhite       byte    "B-WHITE", 0

  s_Yes         byte    "YES", 0
  s_No          byte    "NO", 0
  s_GoodByte    byte    "BYE", 0


pub check_commands | idx

  if (CMD_MODE == M_USB)                                        ' echo if using terminal
    cmd.str(tbuf.str)
    cmd.tx(13)

  idx := tbuf.str_index(@Commands, 23)                          ' get index of command

  if (idx == -1)                                                ' not found
    return false

  case idx
    0  : tcolor := OB_Red
    1  : tcolor := OB_Orange
    2  : tcolor := OB_Yellow
    3  : tcolor := OB_Green
    4  : tcolor := OB_Cyan
    5  : tcolor := OB_Blue
    6  : tcolor := OB_Violet
    7  : tcolor := OB_Purple
    8  : tcolor := OB_Pink
    9  : tcolor := OB_White

    10 : set_board(OB_Red)
    11 : set_board(OB_Orange)
    12 : set_board(OB_Yellow)
    13 : set_board(OB_Green)
    14 : set_board(OB_Cyan)
    15 : set_board(OB_Blue)
    16 : set_board(OB_Violet)
    17 : set_board(OB_Purple)
    18 : set_board(OB_Pink)
    19 : set_board(OB_White)

    20 : answer_yes
    21 : answer_no
    22 : say_goodbye

  tbuf.clear                                                    ' clear buffer to stop spelling


pub set_board(newcolor) | phase

  repeat phase from 1 to 255
    strip.set_all(strip.morph(bcolor, newcolor, phase))
    time.pause(4)

  bcolor := newcolor


pub answer_yes

  fade_down_group(0, STRIP_LEN-1, bcolor, FADE_DN_MS)
  time.pause(500)
  fade_up(OB_Yes, OB_Green, FADE_UP_MS)
  fade_down(OB_Yes, OB_Green, FADE_DN_MS)
  time.pause(1000)
  fade_up_group(0, STRIP_LEN-1, bcolor, FADE_UP_MS)


pub answer_no

  fade_down_group(0, STRIP_LEN-1, bcolor, FADE_DN_MS)
  time.pause(500)
  fade_up(OB_No, OB_Red, FADE_UP_MS)
  fade_down(OB_No, OB_Red, FADE_DN_MS)
  time.pause(1000)
  fade_up_group(0, STRIP_LEN-1, bcolor, FADE_UP_MS)


pub say_goodbye

  fade_down_group(0, STRIP_LEN-1, bcolor, FADE_DN_MS)
  time.pause(500)
  fade_up_group(OB_GoodbyeLo, OB_GoodbyeHi, OB_White, FADE_UP_MS)
  fade_down_group(OB_GoodbyeLo, OB_GoodbyeHi, OB_White, FADE_DN_MS)
  time.pause(1000)
  fade_up_group(0, STRIP_LEN-1, bcolor, FADE_UP_MS)


pub show_string | len, idx, c

  len := tbuf.size

  repeat idx from 0 to len-1
    c := buffer[idx]
    case c
      "A".."Z" : show_letter(c)
      "0".."9" : show_number(c)
      " "      : time.pause(FADE_DN_MS)


pub show_letter(c)

  c := Alphas[c - "A"]
  fade_up(c, tcolor, FADE_UP_MS)
  fade_down(c, tcolor, FADE_DN_MS)


pub show_number(c)

  c := Numbers[c - "0"]
  fade_up(c, tcolor, FADE_UP_MS)
  fade_down(c, tcolor, FADE_DN_MS)


pub setup

'' Configure HC-8+ IO and support objects

  time.start                                                    ' setup for timing/delays

  prng.seed(-cnt, -cnt ~> 2, $EF8_7E4, cnt <- 2, cnt)           ' seed randomizer

  tbuf.use(@buffer, BUF_SIZE)                                   ' configure input text buffer
  tbuf.clear

  io.start(0, 0)                                                ' clear pins (this cog)

  outs.start(CHANNELS, OUT0)                                    ' start outputs driver *

  strip.start_2811(@pixels, STRIP_LEN, OUT7, 0_5)               ' start WS2811 pixel driver *

  set_rs485(NOCON)                                              ' RS-485 not connected

  if (CMD_MODE == M_USB)
    cmd.start(RX1, TX1, %0000, TRM_BAUD)                        ' use PST *
  else
    cmd.start(RX3, TX3, %0000, BLE_BAUD)                        ' usb BT/BLE *

  bcog := cognew(background, @bcstack) + 1                      ' start background cog *


con

  { --------------------------------- }
  {  S U P P O R T   R O U T I N E S  }
  { --------------------------------- }


pub ttl_pin

'' Reads TTL states for 1-of-8 input
'' -- returns -1 (no input)
'' -- returns 0 to 7 for input on IN0 to IN7
''    * only one input allowed

  if (bit_count(ttlpins) == 1)                                  ' only one button pressed?
    return bit_pos(ttlpins, 0)                                  '  yes, return button #
  else
    return -1                                                   '  none or bad input


pub ttl_inputs

'' Return TTL inputs status
'' -- pins are scanned by background cog

  return ttlpins


pub dmx_address

'' Return DMX address switch setting
'' -- switches are scanned by background cog

  return dmxaddr


pub read_xport | xp                                             ' Rev F and higher only

'' Returns state of xport bits
'' -- does not set dira bits to 0

  if (PCB_REV < "F")
    return %0000
  else
    return (ina[XP3] << 3) | (ina[XP2] << 2) | (ina[XP1] << 1) | ina[XP0]


pub write_xport(value, outsmask) | idx, mask                    ' Rev F and higher only

'' Write value to xport pins
'' -- outsmask used to set xport output pins
'' -- will not change pins with 0 bit in outsmask

  if (PCB_REV < "F")
    return

  ' unrolled

  if (outsmask & %1000)
    io.set(XP3, value >> 3)

  if (outsmask & %0100)
    io.set(XP2, value >> 2)

  if (outsmask & %0010)
    io.set(XP1, value >> 1)

  if (outsmask & %0001)
    io.set(XP0, value >> 0)


pub set_rgy_solid(color)

'' Sets R/G LED
'' -- direct control only
'' -- red/green/yellow + blinking requires support cog

  ifnot (bcog)                                                  ' if support cog not running
    dira[R_LED..G_LED] := %11                                   ' make outputs
    case color                                                  ' set color to state
      OFF : outa[R_LED..G_LED] := %00
      RED : outa[R_LED..G_LED] := %10
      GRN : outa[R_LED..G_LED] := %01
      YEL : outa[R_LED..G_LED] := %00                           ' yellow requires active driver

  else
    case color
      OFF : set_rgy_flash(OFF, 100, OFF, 100)                   ' set to solid color
      RED : set_rgy_flash(RED, 100, RED, 100)
      GRN : set_rgy_flash(GRN, 100, GRN, 100)
      YEL : set_rgy_flash(YEL, 100, YEL, 100)


pub set_rgy_flash(p0c, p0ms, p1c, p1ms)

'' Explicit (control registers) setup of OGRY LED
'' -- requires support cog to be running
'' -- p0c is phase 0 color (OFF, GRN, RED, YEL)
'' -- p0ms is phase 0 timing in milliseconds
'' -- p1c is phase 1 color
'' -- p1ms is phase 1 timing in milliseconds

  ifnot (bcog)
    return

  if ((p0c => OFF) and (p0c =< YEL))                            ' phase 0 color
    rgycolor[0] := p0c
  else
    rgycolor[0] := OFF

  if (p0ms > 0)                                                 ' phase 0 timing
    rgytime[0] := p0ms
  else
    rgytime[0] := 100

  if ((p1c => OFF) and (p1c =< YEL))                            ' phase 1 color
    rgycolor[1] := p1c
  else
    rgycolor[1] := OFF

  if (p1ms > 0)                                                 ' phase 1 timing
    rgytime[1] := p1ms
  else
    rgytime[1] := 100


pub set_rs485(state)

'' Sets RS-485 RX and TX enable inputs

  case state
    NOCON : outa[RX2_EN..TX2_EN] := %10
    RX    : outa[RX2_EN..TX2_EN] := %00
    TX    : outa[RX2_EN..TX2_EN] := %11
    RXTX  : outa[RX2_EN..TX2_EN] := %01

  dira[RX2_EN..TX2_EN] := %11


pub bit_count(value) : bc

'' Returns # of bits set (1) in value

  repeat while (value)                                          ' count while there are bits
    bc += (value & %1)                                          ' add lsb bit value to count
    value >>= 1                                                 ' next bit


pub bit_pos(value, mode)

'' Returns position of 1st bit set (0..31)
'' -- mode 0 to scan from lsb, mode 1 to scan from msb
'' -- -1 = no bits set

  if (value == 0)                                               ' if no bits
    return -1                                                   '  return -1
  else
    if (mode == LSBFIRST)                                       ' check from LSB
      value ><= 32                                              '  flip for >|
      return (32 - >|value)
    else
      return (>|value - 1)


pub bit_val(value, pos)

'' Returns bit (0..1) from pos (0..31) in value

  if ((pos => 0) and (pos =< 31))
    return (value >> pos) & 1
  else
    return 0


con

  { ----------------------------- }
  {  B A C K G R O U N D   C O G  }
  {  - global milliseconds        }
  {  - R/G LED                    }
  {  - TTL/DMX inputs scanning    }
  { ----------------------------- }


var

  long  bcog                                                    ' background cog #
  long  bcstack[64]                                             ' stack for background cog

  long  millis                                                  ' global milliseconds register


pri background | t                                              ' launch with cognew()

  io.low(R_LED)                                                 ' setup R/G LED pins
  io.low(G_LED)

  io.high(LD_165)                                               ' setup x165 io pins
  io.low(CLK_165)
  io.input(DO_165)

  millis := 0

  t := cnt                                                      ' sync loop timer
  repeat
    waitcnt(t += MS_001)                                        ' run loop every millisecond
    ++millis
    refresh_rg_led                                              ' update the R/G LED
    scan_ttl_ins                                                ' re-scan TTL & DMX inputs


var

  long  rgycolor[2]                                             ' led phase colors
  long  rgytime[2]                                              ' led phase timing (ms)
  long  rgphase                                                 ' current phase (red or green chip)
  long  phasetimer                                              ' time in phase
  long  rgtimer                                                 ' timer for rg process


pri refresh_rg_led                                              ' call only from support()

  if (++phasetimer => rgytime[rgphase])                         ' done with this phase?
    phasetimer := 0                                             ' yes, reset timer
    rgphase := 1 - rgphase                                      '  and invert phase

  if (++rgtimer == 16)
    rgtimer := 0

  case rgycolor[rgphase]                                        ' set led to color for phase
    OFF:
      outa[R_LED..G_LED] := %00

    GRN:
      outa[R_LED..G_LED] := %01

    RED:
      outa[R_LED..G_LED] := %10

    YEL:
      if (rgtimer < 2)
        outa[R_LED..G_LED] := %10                               ' 1 red
      else
        outa[R_LED..G_LED] := %01                               ' 15


var

  long  ttlpins                                                 ' updated by background cog
  long  dmxaddr


pri scan_ttl_ins | tempin

'' Scan TTL and DMX address inputs

  outa[LD_165] := 0                                             ' blip Shift/Load line
  outa[LD_165] := 1

  ' unrolled for best speed

  tempin := (tempin << 1) | ina[DO_165]                         ' bit15
  outa[CLK_165] := 1                                            ' blip clock
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0

  tempin := (tempin << 1) | ina[DO_165]                         ' bit7
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165]                         ' bit0

  ttlpins := tempin.byte[0]                                     ' update global vars
  dmxaddr := tempin.byte[1] | (ina[DMX_A8] << 8)


dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}