'' =================================================================================================
''
''   File....... jm_hm-10_connect.spin
''   Purpose....
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2017 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started....
''   Updated.... 24 OCT 2017
''
'' =================================================================================================

{

  Product:
  -- https://www.amazon.com/gp/product/B06WGZB2N4

  AT Commands (overview):
  -- http://fab.cba.mit.edu/classes/863.15/doc/tutorials/programming/bluetooth.html

  Docs:
  -- http://fab.cba.mit.edu/classes/863.15/doc/tutorials/programming/bluetooth/bluetooth40_en.pdf

}


con { timing }

  _clkmode = xtal1 + pll16x                                     
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us

  TRM_BAUD = 115_200                                            ' terminal baud rate
  BLE_BAUD =   9_600                                            ' ble baud rate
  

con { io pins }                                                  
                                                                 
  RX1  = 31  { I }                                              ' serial / programming
  TX1  = 30  { O }                                            
                                                              
  SDA1 = 29  { I/O }                                            ' i2c / eeprom
  SCL1 = 28  { I/O } 

  RX2  =  5  { I }                                              ' ble
  TX2  =  4  { O }

   
con

  #true, ON, OFF                                                ' digital control

  #0, LSBFIRST, MSBFIRST

  BUF_SIZE = 40   


obj

' main                                                          ' * master Spin cog
  time : "jm_time_80"                                           '   timing and delays
  io   : "jm_io_basic"                                          '   essential io
  tbuf : "jm_textbuf"                                           '   text buffer management
  term : "jm_fullduplexserial"                                  ' * serial IO for terminal
  ble  : "jm_fullduplexserial"                                  ' * serial IO for BLE module    
                                                                 
' * uses cog when loaded                                         
                                                                 
                                                                 
var

  byte  buffer[BUF_SIZE+1]
                                                                 
                                                                 
pub main                                                      
                                                                 
  setup

  repeat
    get_command
    show_response    


pub get_command : c

  repeat
    c := term.rxcheck                                           ' character waiting?
    if (c < 0)                                                  ' no, abort
      return
    else
      if (c <> 13)                                              ' if c is not CR
        if (c == 8)                                             ' backspace?
          tbuf.backspace                                        '  delete last character
        else                                                    ' else
          tbuf.putc(tbuf.upper(c))                              '  put it into buffer
      else
        if (tbuf.size > 0)                                      ' anything actually in the buffer?                                             
          term.str(string(13, "--> "))                          ' display it         
          term.str(tbuf.str)    
          term.tx(13)               
          ble.str(tbuf.str)                                     ' send to ble   
          ble.txflush                                           ' let command finish            
          tbuf.clear                                            ' clean-up for next command              
                                                                 

pub show_response : c

  if (ble.available)                                            ' anything waiting?
    term.str(string("<-- "))                                    ' yes, show response prompt   
    repeat
      c := ble.rxtime(2)                                        ' get chars
      if (c => 32)                                              ' if valid
        term.tx(c)                                              '  display
      else
        term.tx(13)
        return

                                                                                
pub setup                                                        
                                                                 
'' Setup IO and objects for application                          
                                                                 
  time.start                                                    ' setup timing & delays

  io.start(0, 0)                                                ' clear all pins (master cog)

  tbuf.use(@buffer, BUF_SIZE)                                   ' setup text buffer
  tbuf.clear
                                                              
  term.start(RX1, TX1, %0000, TRM_BAUD)                         ' start serial for terminal

  ble.start(RX2, TX2, %0000, BLE_BAUD)                          ' start serial for BLE

  time.pause(10)
  term.rxflush
  ble.rxflush  
  
                          
dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      